# Programmable logic block functions
from .PLB import *

# Input/output functions
from .IOB import *

# Block RAM functions
from .RAMB import *

# PLL
from .PLL import *

FAMILY = 'ice40'

print('import lattice ice40')
